#include "physical_resource_manager.h"
#include "device.h"


static VmaAllocator CreateAllocator(rt_vk_device *dev) {
#define SET_FNC(name)     fncs.name = name
#define SET_KHR_FNC(name) (fncs).name##KHR = name
    VmaVulkanFunctions fncs                = {NULL};
    SET_FNC(vkGetInstanceProcAddr);
    SET_FNC(vkGetDeviceProcAddr);
    SET_FNC(vkGetPhysicalDeviceProperties);
    SET_FNC(vkGetPhysicalDeviceMemoryProperties);
    SET_FNC(vkAllocateMemory);
    SET_FNC(vkFreeMemory);
    SET_FNC(vkMapMemory);
    SET_FNC(vkUnmapMemory);
    SET_FNC(vkFlushMappedMemoryRanges);
    SET_FNC(vkInvalidateMappedMemoryRanges);
    SET_FNC(vkBindBufferMemory);
    SET_FNC(vkBindImageMemory);
    SET_FNC(vkGetBufferMemoryRequirements);
    SET_FNC(vkGetImageMemoryRequirements);
    SET_FNC(vkCreateBuffer);
    SET_FNC(vkDestroyBuffer);
    SET_FNC(vkCreateImage);
    SET_FNC(vkDestroyImage);
    SET_FNC(vkCmdCopyBuffer);
    SET_KHR_FNC(vkGetBufferMemoryRequirements2);
    SET_KHR_FNC(vkGetImageMemoryRequirements2);
    SET_KHR_FNC(vkBindBufferMemory2);
    SET_KHR_FNC(vkBindImageMemory2);
    SET_KHR_FNC(vkGetPhysicalDeviceMemoryProperties2);
    SET_FNC(vkGetDeviceBufferMemoryRequirements);
    SET_FNC(vkGetDeviceImageMemoryRequirements);
#undef SET_FNC
#undef SET_KHR_FNC

    VmaAllocatorCreateInfo allocator_info = {
        .instance             = dev->instance,
        .physicalDevice       = dev->phys_device,
        .device               = dev->device,
        .pAllocationCallbacks = dev->alloc_cb,
        .vulkanApiVersion     = VK_API_VERSION_1_3,
        .pVulkanFunctions     = &fncs,
    };

    VmaAllocator allocator;
    if (vmaCreateAllocator(&allocator_info, &allocator) != VK_SUCCESS)
        return NULL;
    return allocator;
}

rt_create_vk_physical_resource_manager_result rtCreateVkPhysicalResourceManager(struct rt_vk_device *dev) {
    rt_vk_physical_resource_manager phys_res_mgr;
    phys_res_mgr.dev = dev;
    phys_res_mgr.allocator = CreateAllocator(dev);
    if (!phys_res_mgr.allocator) {
        return (rt_create_vk_physical_resource_manager_result){.result =  RT_UNKNOWN_ERROR};
    }

    return (rt_create_vk_physical_resource_manager_result){.result =  RT_SUCCESS, .phys_res_mgr = phys_res_mgr};
}

void rtDestroyVkPhysicalResourceManager(rt_vk_physical_resource_manager *phys_res_mgr) {
    vmaDestroyAllocator(phys_res_mgr->allocator);
}